import pandas as pd
import os


class BaseZRP():
    """
    Prepares data to generate race & ethnicity proxies
    
    Parameters
    ----------

    support_files_path: "data/processed"
        File path with support data.
    key: str, default 'ZEST_KEY'
        Key to set as index. If not provided, a key will be generated.
    first_name: str, default 'first_name'
        Name of first name column
    middle_name: str, default 'middle_name'
        Name of middle name column
    last_name: str, default 'last_name'
        Name of last name/surname column
    house_number: str, default 'house_number'
        Name of house number column. Also known as primary address number this is the unique number assigned to a building to delineate it from others on a street. This is usually the first component of a delivery address line.
    street_address: str, default 'street_address'
        Name of street address column. The street address is usually comprised of predirectional, street name, and street suffix. 
    city: str, default 'city'
        Name of city column.
    state: str, default 'state'
        Name of state column.
    zip_code: str, default 'zip_code'
        Name of zip or postal code column.
    race: str, default 'race'
        Name of the race column
    census_tract: str
        Name of census tract column
    street_address_2: str, optional
        Name of additional address column
    name_prefix: str, optional
        Name of column containing full name preix (ie Dr, Sr, and Esq )
    name_suffix: str, optional
        Name of column containing full name suffix (ie jr, iii, and phd)
    na_values: list, optional
        List of missing values to replace.
    file_path: str, optional
        Path where to put artifacts and other files generated during intermediate steps. 
    geocode: bool, default True
        Whether to geocode.
    bisg: bool, default True
        Whether to return BISG proxies. 
    readout: bool, default True
        Whether to return a readout.
    n_jobs: int, default 49
        Number of jobs in parallel
    year: str, default '2019'
        ACS year to use.
    span: str, default '5'
        Year span of ACS data to use.
    runname: str, default 'test'
    """

    def __init__(self, support_files_path="data/processed", key="ZEST_KEY", first_name="first_name",
                 middle_name="middle_name", last_name="last_name", house_number="house_number",
                 street_address="street_address", city="city", state="state", zip_code="zip_code", race='race',
                 census_tract=None, block_group=None, street_address_2=None, name_prefix=None, name_suffix=None,
                 na_values=None, file_path=None, geocode=True, bisg=True, readout=True, n_jobs=49, year="2019",
                 span="5", runname="test"):
        self.key = key
        self.first_name = first_name
        self.middle_name = middle_name
        self.last_name = last_name
        self.name_suffix = name_suffix
        self.house_number = house_number
        self.street_address = street_address
        self.street_address_2 = street_address_2
        self.city = city
        self.state = state
        self.zip_code = zip_code
        self.census_tract = census_tract
        self.block_group = block_group
        self.file_path = file_path
        self.support_files_path = support_files_path
        self.na_values = na_values
        self.geocode = geocode
        self.readout = readout
        self.bisg = bisg
        self.race = race
        self.n_jobs = n_jobs
        self.year = year
        self.span = span
        self.runname = runname
        if file_path:
            self.out_path = os.path.join(self.file_path, "artifacts")
        else:
            self.out_path = "artifacts"

        super().__init__()

    def fit(self):
        pass

    def transform(self):
        pass



if __name__ == '__main__':
    from .download import download
    import plac
    import sys

    commands = {
        "download": download,
    }

    if len(sys.argv) == 1:
        print("Available commands:", ", ".join(commands))
        sys.exit(1)

    command = sys.argv.pop(1)

    if command in commands:
        plac.call(commands[command], sys.argv[1:])
    else:
        print("Unknown command:", command)
        print("Available commands:", ", ".join(commands))
        sys.exit(1)


from pycm import ConfusionMatrix

from zrp.prepare.utils import load_file
from zrp.prepare.preprocessing import set_id
from pycm import ConfusionMatrix
import pandas as pd
import numpy as np
import os
import re


class ZRP_Performance():
    """
    Generates performance analysis artifacts
    
    Parameters
    ----------
    key: str 
        Key to set as index. If not provided, a key will be generated.
    race: str
        Name of race column 
    proxy_data_path: str
        File path to proxy data
    ground_truth_path: str
        File path to ground truth data
    """

    def __init__(self, proxy_data_path=None,
                 ground_truth_path=None, key="ZEST_KEY", race="race"):
        self.key = key
        self.proxy_data_path = proxy_data_path
        self.ground_truth_path = ground_truth_path
        self.race = race

    def fit(self):
        return self

    def transform(self, proxy_data = None, ground_truth = None):
        """
        Returns confusion matrix analysis of ZRP performance against gound truth in the form of a dictionary

        Parameters
        ----------
        proxy_data: pd.dataframe
            Dataframe containing proxy race labels
        ground_truth: pd.dataframe
            Dataframe containing ground truth race labels
            """
        # Load Data
        try:
            proxies = proxy_data.copy()
        except AttributeError:
            proxies = load_file(self.proxy_data_path)
        try:
            ground_truth = ground_truth.copy()
        except AttributeError:
            ground_truth = load_file(self.ground_truth_path)

        proxies = set_id(proxies, self.key)
        ground_truth = set_id(ground_truth, self.key)
        proxies = proxies[f"{self.race}_proxy"]
        ground_truth = ground_truth[self.race]

        cm = ConfusionMatrix(
            np.array(ground_truth),
            np.array(proxies)
        )
        performance_dict = {}

        for metric in ["PPV", "TPR", "FPR", "FNR", "TNR", "AUC"]:
            performance_dict[metric] = None

        performance_dict['PPV'] = cm.PPV
        performance_dict['TPR'] = cm.TPR
        performance_dict['FPR'] = cm.FPR
        performance_dict['FNR'] = cm.FNR
        performance_dict['TNR'] = cm.TNR
        performance_dict['AUC'] = cm.AUC

        return(performance_dict)



import numpy as np
import pandas as pd
import re
from sklearn.base import BaseEstimator, TransformMixin

class ZRPEngineeringBase(BaseEstimator, TransformMixin):
    """Base class for feature engineering"""
    def __init__(self, *args, **kwargs):
        self.zrp = True
    def fit(self, X, y):
        return self
    def transform(self, X):
        pass


import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.rst").read_text()

# The contents of the requirements.txt
with open("requirements.txt", 'r') as file:
    requirements_raw = file.readlines()

requirements = []
for r in requirements_raw:
    requirements.append(r.replace('\n', ''))

# This call to setup() does all the work
setup(
    name="zrp",
    version="0.1.19",
    description="The Zest Race Predictor tool predicts race/ethnicity using a name and address as inputs.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/zestai/zrp",
    author="Kasey Matthews et al.",
    author_email="abetterway@zest.ai",
    license="Apache-2.0",
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 4 - Beta",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
    ],
    keywords="race ethnicity names address acs geocode",
    packages=find_packages(exclude=("extras")),
    package_data={
        "zrp": [
        ],
    },
    include_package_data=True,
    install_requires=requirements,  # TODO: put back requirements variable
    entry_points={
        "console_scripts": [

            # TODO: Once we know what endpoints are being exposed (see ethnicolr's as example)
        ]
    },
)  # TODO: Update Manifest.in file to include data and other files as suggested here: https://realpython.com/pypi-publish-python-package/#adding-files-to-your-package



from .preprocessing import *
from .base import BaseZRP
from .utils import *
import pandas as pd
import os


def acs_search(year, span):
    """
    Searches for processed ACS data
    
    Parameters:
    -----------
    year: str
        Release year of ACS data
    span: str
        Span of ACS data (ie '1' or '5')      
    """
    file_list_z = []
    file_list_c = []
    file_list_b = []
    curpath = dirname(__file__)
    data_path = join(curpath, f'../data/processed/acs/{year}/{span}yr')
    for root, dirs, files in os.walk(os.path.join(data_path)):
        for file in files:
            if (f"_zip" in file) & ("processed" in file):
                file_list_z.append(os.path.join(root, file))
            if (f"tract" in file) & ("processed" in file):
                file_list_c.append(os.path.join(root, file))
            if (f"blockgroup" in file) & ("processed" in file):
                file_list_b.append(os.path.join(root, file))
    return (file_list_z, file_list_c, file_list_b)


class ACSModelPrep(BaseZRP):
    """
    Prepares ACS data & processed user input for modeling
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def fit(self):
        pass

    def acs_combine(self, data, acs_bg, acs_ct, acs_zip):
        """
        Combines ACS data with processed user input data.
        Generating optional features for modeling.
        
        Parameters
        ----------
        data: str
            Processed user input data, expected to include names & GEOID        
        acs_bg: str
            ACS block group lookup table
        acs_ct: str
            ACS census tract lookup table
        acs_zip: str
            ACS zip code lookup table
        """
        # Block Group
        mbggk_list = list(set(data.GEOID_BG.unique()).intersection(set(acs_bg.GEOID.unique())))

        print(" ...Copy dataframes")
        ## Merge by current
        mbggk = data.copy()
        mctgk = data.copy()
        mbz = data.copy()
        nm = data.copy()

        print(" ...Block group")

        mbggk = mbggk[(mbggk.GEOID_BG.isin(mbggk_list))].reset_index(drop=False)
        mbggk_zkeys = list(mbggk.index)
        mbggk = mbggk.merge(acs_bg,
                            left_on="GEOID_BG",
                            right_on="GEOID").set_index(self.key)
        mbggk["acs_source"] = "BG"

        # Census Tract
        mctgk_list = list(set(data.GEOID_CT.unique()).intersection(set(acs_ct.GEOID.unique())))
        prev_zkeys = mbggk_zkeys

        mctgk = mctgk[~(mctgk.index.isin(prev_zkeys)) & \
                      (mctgk["GEOID_CT"].isin(mctgk_list))].reset_index(drop=False)
        mctgk_zkeys = list(mctgk.index)
        print(" ...Census tract")
        mctgk = mctgk.merge(acs_ct,
                            left_on="GEOID_CT",
                            right_on="GEOID").set_index(self.key)
        mctgk["acs_source"] = "CT"

        # Merge by Zip
        mbz_list = list(set(data["GEOID_ZIP"].unique()).intersection(set(acs_zip.GEOID.unique())))
        prev_zkeys_0 = mbggk_zkeys + mctgk_zkeys
        print(" ...Zip code")

        mbz = mbz[~(mbz.index.isin(prev_zkeys_0)) & \
                  (mbz["GEOID_ZIP"].isin(mbz_list))].reset_index(drop=False)
        mbz_zkeys = list(mbz.index)
        mbz = mbz.merge(acs_zip,
                        right_on="GEOID",
                        left_on="GEOID_ZIP").set_index(self.key)
        mbz["acs_source"] = "ZIP"

        # No Merge
        print(" ...No match")
        prev_zkeys_1 = mbggk_zkeys + mctgk_zkeys + mbz_zkeys
        nm = nm[~(nm.index.isin(prev_zkeys_1))]
        nm["acs_source"] = None

        print(" ...Merge")
        data_out = pd.concat([mbggk, mctgk, mbz], sort=True)
        data_out = pd.concat([data_out, nm], sort=True)
        print(" ...Merging complete")
        return (data_out)

    def transform(self, input_data, save_table=False):
        """
        Parameters
        ----------
        input_data: pd.Dataframe
            Dataframe to be transformed
        save_table: bool
            Optional save
        """
        # Load Data
        try:
            data = input_data.copy()
        except AttributeError:
            data = load_file(input_data)
            print("Input data file is loaded")
        
        data = data.rename(columns = {self.first_name : "first_name", 
                                      self.middle_name : "middle_name", 
                                      self.last_name : "last_name",
                                      self.house_number : "house_number", 
                                      self.street_address : "street_address", 
                                      self.city : "city",
                                      self.zip_code : "zip_code",
                                      self.state : "state", 
                                      self.block_group : "block_group", 
                                      self.census_tract : "census_tract"
                             }
                  )

        if "GEOID_ZIP" not in data.columns:
            print("Generating Geo IDs")
            data["GEOID_ZIP"] = data["ZEST_ZIP"]
            data["GEOID_CT"] = data["STATEFP"] + data["COUNTYFP"] + data["TRACTCE"]
            data["GEOID_BG"] = data["GEOID_CT"] + data["BLKGRPCE"]

        file_list_z, file_list_c, file_list_b = acs_search(self.year,
                                                           self.span)

        print("   ...loading ACS lookup tables")
        acs_bg = load_file(file_list_b[0])
        acs_ct = load_file(file_list_c[0])
        acs_zip = load_file(file_list_z[0])

        print("   ... combining ACS & user input data")
        data_out = self.acs_combine(data,
                                    acs_bg,
                                    acs_ct,
                                    acs_zip)

        if save_table:
            make_directory(self.out_path)
            file_name = f"Zest_processed_data_.parquet"
            save_dataframe(data_out, self.out_path, file_name)
        return (data_out)



from zrp.prepare.preprocessing import *
from os.path import dirname, join, expanduser
from zrp.prepare.utils import *
import pandas as pd
import numpy as np
import fiona
import json
import sys
import os
import re


def sides_split(addr_edge_df):
    """Split tigerline dataframes into sides
    Parameters
    ----------
    addr_edge_df: pd.DataFrame/tigerline dataframe with address features & edges data
    """
    l_addr_edge_df = addr_edge_df.copy()
    r_addr_edge_df = addr_edge_df.copy()

    l_addr_edge_df = l_addr_edge_df.drop(['TFIDR', 'ARIDR', 'RFROMHN',
                                           'RTOHN', 'ZIPR', 'PARITYR',
                                           'PLUS4R', 'RFROMTYP', 'RTOTYP',
                                           'OFFSETR', 'RFROMADD', 'RTOADD'],
                                         axis=1)
    r_addr_edge_df = r_addr_edge_df.drop(['TFIDL', 'ARIDL', 'LFROMHN', 
                                          'LTOHN', 'ZIPL', 'PARITYL',
                                          'PLUS4L', 'LFROMTYP', 'LTOTYP',
                                          'OFFSETL', 'LFROMADD', 'LTOADD'],
                                         axis=1)

    l_addr_edge_df["SIDE"] = "L"
    r_addr_edge_df["SIDE"] = "R"

    l_addr_edge_df = l_addr_edge_df.rename(
        columns={"TFIDL": "TFID",
                 "ARIDL": "ARID",
                 "LFROMHN": "FROMHN",
                 "LTOHN": "TOHN",
                 "ZIPL": "ZIP",
                 "PARITYL": "PARITY",
                 "PLUS4L": "PLUS4",
                 "LFROMTYP":"FROMTYP",
                 "LTOTYP": "TOTYP",
                 "OFFSETL": "OFFSET",
                 "LFROMADD": "FROMADD",
                 "LTOADD": "TOADD" }
    )
    r_addr_edge_df = r_addr_edge_df.rename(
        columns={"TFIDR": "TFID",
                 "ARIDR": "ARID",
                 "RFROMHN": "FROMHN",
                 "RTOHN": "TOHN",
                 "ZIPR": "ZIP",
                 "PARITYR": "PARITY",
                 "PLUS4R": "PLUS4",
                 "RFROMTYP":"FROMTYP",
                 "RTOTYP": "TOTYP",
                 "OFFSETR": "OFFSET",
                 "RFROMADD": "FROMADD",
                 "RTOADD": "TOADD"}
    )
    return (l_addr_edge_df, r_addr_edge_df)


def concat_shapes(df_list):
    """Concatenate lists of shapefile data"""
    df = pd.concat(df_list, axis=0)
    return (df)


def merge_shapes(df_1, df_2, merge_cols, merge_method):
    """Merge shapefile dataframes"""
    df = df_1.merge(df_2, on=merge_cols, how=merge_method)
    return (df)


class GeoLookUpBuilder():
    """
    This class builds geo lookup tables.
    
    Parameters
    ----------
    support_files_path: str
        Path to support files
    year: str (default 2019)
        Year associated with ACS data. 
    """

    def __init__(self, support_files_path, year):
        self.support_files_path = support_files_path
        self.year = year
        self.raw_geo_path = os.path.join(self.support_files_path, "raw/geo", self.year)
        self.out_geo_path = os.path.join(self.support_files_path, "processed/geo", f"{self.year}_backup") #new change update 

    def fit(self):
        return self

    def transform(self, st_cty_code, save_table=True):
        """
        Returns DataFrame of geo lookup tables

        Parameters
        ----------
        st_cty_code: str
            state county code string
        save_table: bool
            Indicator to save dataframe
        """
        curpath = dirname(__file__)
        make_directory(output_directory=self.raw_geo_path)
        make_directory(output_directory=self.out_geo_path)

        geo_support_files_path = self.raw_geo_path
        print("Shapefile input:", geo_support_files_path)
        print("Lookup Table output:", self.out_geo_path)
        print("")

        addrfeat = "".join([geo_support_files_path,
                            "/tl_",
                            self.year,
                            "_",
                            st_cty_code,
                            "_addrfeat.shp"])
        faces = "".join([geo_support_files_path,
                         "/tl_",
                         self.year,
                         "_",
                         st_cty_code,
                         "_faces.shp"])
        edges = "".join([geo_support_files_path,
                         "/tl_",
                         self.year,
                         "_",
                         st_cty_code,
                         "_edges.shp"])

        print(" ... Loading requirements")
        af_df = gdbToDf_short(addrfeat,
                              None)
        fc_df = gdbToDf_short(faces,
                              "face")
        ed_df = gdbToDf_short(edges,
                              "edge")

        af_col_keep = [
            "TLID",
            "TFIDL",
            "TFIDR",
            "ARIDL",
            "ARIDR",
            "LINEARID",
            "FULLNAME",
            "LFROMHN",
            "LTOHN",
            "RFROMHN",
            "RTOHN",
            "ZIPL",
            "ZIPR",
            "EDGE_MTFCC",
            "ROAD_MTFCC",
            "PARITYL",
            "PARITYR",
            "LFROMTYP",
            "LTOTYP",
            "RFROMTYP",
            "RTOTYP",
            "OFFSETL",
            "OFFSETR",
            "PLUS4R",
            "PLUS4L"
        ]

        if "STATEFP20" in fc_df.columns:
            fc_df = fc_df.rename(columns={"STATEFP20": "STATEFP",
                                          "COUNTYFP20": "COUNTYFP",
                                          "TRACTCE20": "TRACTCE",
                                          "BLKGRPCE20": "BLKGRPCE",
                                          "BLOCKCE20": "BLOCKCE",
                                          "TTRACTCE20": "TTRACTCE",
                                          "TBLKGPCE20": "TBLKGPCE",
                                          "ZCTA5CE20": "ZCTA5CE",
                                          "PUMA5CE20": "PUMACE",
                                          "PUMA5CE10": "PUMACE10"
                                          }
                                 )
        else:
            fc_df[["STATEFP", "COUNTYFP", "TRACTCE", "BLKGRPCE", "BLOCKCE", "ZCTA5CE", "PUMACE"]] = fc_df[
                ["STATEFP10", "COUNTYFP10", "TRACTCE10", "BLKGRPCE10", "BLOCKCE10", "ZCTA5CE10", "PUMACE10"]]

        fc_col_keep = [
            "TFID",
            "STATEFP10",
            "COUNTYFP10",
            "TRACTCE10",
            "BLKGRPCE10",
            "BLOCKCE10",
            "ZCTA5CE10",
            "PUMACE10",
            "STATEFP",
            "COUNTYFP",
            "TRACTCE",
            "BLKGRPCE",
            "BLOCKCE",
            "ZCTA5CE",
            "TTRACTCE",
            "TBLKGPCE",
            "OFFSET",
            "PUMACE"
        ]

        ed_col_keep = [
            "STATEFP",
            "COUNTYFP",
            "TLID",
            "TFIDL",
            "TFIDR",
            "FULLNAME",
            "LFROMADD",
            "LTOADD",
            "RFROMADD",
            "RTOADD",
            "ZIPL",
            "ZIPR",
            "OFFSETL",
            "OFFSETR",
        ]

        af_df = af_df[af_col_keep]
        fc_df = fc_df[fc_col_keep]
        ed_df = ed_df[ed_col_keep]

        print(" ... Creating lookup table")

        addr_edge_df = merge_shapes(af_df,
                                    ed_df,
                                    ["TLID", "TFIDL", "TFIDR",
                                     "ZIPL", "ZIPR", "FULLNAME",
                                     "OFFSETL", "OFFSETR"],
                                    "inner")
        
        l_addr_edge_df, r_addr_edge_df = sides_split(addr_edge_df)

        l_addr_edge_face_df = merge_shapes(l_addr_edge_df,
                                           fc_df,
                                           ["TFID", "COUNTYFP", "STATEFP", "OFFSET"],
                                           "inner")
        r_addr_edge_face_df = merge_shapes(r_addr_edge_df,
                                           fc_df,
                                           ["TFID", "COUNTYFP", "STATEFP", "OFFSET"],
                                           "inner")
        
        l_addr_edge_face_df = l_addr_edge_face_df[l_addr_edge_face_df["ZIP"].notna()] 
        r_addr_edge_face_df = r_addr_edge_face_df[r_addr_edge_face_df["ZIP"].notna()]         

        aef = concat_shapes([l_addr_edge_face_df,
                             r_addr_edge_face_df])

        aef.drop_duplicates(inplace=True)

        print(" ... Formatting lookup table")
        # rename features to avoid user overlap
        aef = aef.rename(columns={"ZIP": "ZEST_ZIP",
                                  "FULLNAME": "ZEST_FULLNAME"})

        aef["RAW_ZEST_ZIP"] = aef["ZEST_ZIP"].copy()
        aef["RAW_ZEST_STATEFP"] = aef["STATEFP"].copy()
        aef["RAW_ZEST_COUNTYFP"] = aef["COUNTYFP"].copy()
        aef["RAW_ZEST_FULLNAME"] = aef["ZEST_FULLNAME"].copy()
        aef["RAW_ZEST_TRACTCE"] = aef["TRACTCE"].copy()
        aef["RAW_ZEST_BLKGRPCE"] = aef["BLKGRPCE"].copy()

        aef = aef.astype(str)
        gps = ProcessGLookUp()

        data_path = join(curpath, '../data/processed/state_mapping.json')
        state_mapping = load_json(data_path)

        aef = gps.transform(aef, state_mapping)

        # make geo keys/ids
        aef["GEOID_ZIP"] = aef["ZEST_ZIP"]
        aef["GEOID_CT"] = aef["STATEFP"] + aef["COUNTYFP"] + aef["TRACTCE"]
        aef["GEOID_BG"] = aef["GEOID_CT"] + aef["BLKGRPCE"]

        # optional save 
        if save_table:
            make_directory(self.out_geo_path)
            filename = "".join(["Zest_Geo_Lookup_", self.year, "_", st_cty_code, ".parquet"])
            print(f"Saving {filename}")
            save_dataframe(aef, self.out_geo_path, filename)
        else:
            print("No tables were saved")

        return (aef)
